# pyright: reportShadowedImports=false
import math
from typing import *

import numpy as np

from ...common import Detection, Image, Images, InputFormat, clip
from ...runner import Runner, RunnerConfig
from ..onnx import OnnxNodeInfo, OnnxRunner, OnnxRunnerConfig
from .postproc import YoloXPostProcessRunner, YoloXPostProcessRunnerConfig
from .preproc import YoloXPreProcessRunner, YoloXPreProcessRunnerConfig

__all__ = ["YoloXModelRunner", "YoloXModelRunnerConfig"]


class YoloXModelRunnerConfig(
    YoloXPreProcessRunnerConfig,
    OnnxRunnerConfig,
    YoloXPostProcessRunnerConfig,
    dynamic=True,
):
    type: str = "YoloXModelRunner"
    input_node: str = "images"
    output_node: str = "output"
    batch_size: int = -1


class YoloXModelRunner(Runner):
    config_type: Type[RunnerConfig] = YoloXModelRunnerConfig
    outputs: List[str] = ["detections"]

    def prepare(self) -> "YoloXModelRunner":
        config: YoloXModelRunnerConfig = self.config

        self.onnx_runner = OnnxRunner(config)
        self.onnx_runner.prepare()

        self.input_info: OnnxNodeInfo = self.onnx_runner.metadata[config.input_node]

        if config.input_format == InputFormat.hwc:
            (
                config.batch_size,
                config.height,
                config.width,
                _,
            ) = self.input_info.dims

        if config.input_format == InputFormat.chw:
            (
                config.batch_size,
                _,
                config.height,
                config.width,
            ) = self.input_info.dims

        self.output_info: OnnxNodeInfo = self.onnx_runner.metadata[config.output_node]
        _, config.num_anchors, num_features = self.output_info.dims
        config.num_classes = num_features - 5

        self.preproc_runner = YoloXPreProcessRunner(config)
        self.preproc_runner.prepare()

        self.postproc_runner = YoloXPostProcessRunner(config)
        self.postproc_runner.prepare()

        return self

    def run(
        self, images: Union[Image, Images], **kwargs
    ) -> Union[List[Detection], List[List[Detection]]]:
        config: YoloXModelRunnerConfig = self.config

        is_single_image = type(images) is not list
        if is_single_image:
            images = [images]

        images = [
            Image(mat=image) if isinstance(image, np.ndarray) else image
            for image in images
        ]

        preproc_images = []
        for image in images:
            preproc_image = self.preproc_runner.run(image)
            preproc_images.append(preproc_image.mat)

        candidates = []
        for i in range(0, len(preproc_images), config.batch_size):
            batch = preproc_images[i : i + config.batch_size]
            batch_size = len(batch)
            if batch_size < config.batch_size:
                batch += [
                    np.zeros(self.input_info.dims[1:], dtype=self.input_info.dtype)
                    for _ in range(config.batch_size - batch_size)
                ]

            batch = np.stack(batch)
            output = self.onnx_runner.run(**{config.input_node: batch})[
                config.output_node
            ]

            batch = batch[:batch_size, ...]
            output = output[:batch_size, ...]

            candidates.extend(self.postproc_runner.run(output=output))

        results = []
        for candidate, image in zip(candidates, images):
            height, width, _ = image.mat.shape
            top, bottom, left, right = image.padding

            result = []
            for detection in candidate:
                detection.left = clip(detection.left - left, 0, 1)
                detection.top = clip(detection.top - top, 0, 1)
                detection.right = clip(detection.right - right, 0, 1)
                detection.bottom = clip(detection.bottom - bottom, 0, 1)

                if (
                    detection.right <= detection.left
                    or detection.bottom <= detection.top
                ):
                    continue

                detection.left = math.floor(detection.left * width)
                detection.top = math.floor(detection.top * height)
                detection.right = math.floor(detection.right * width)
                detection.bottom = math.floor(detection.bottom * height)

                result.append(detection)

            results.append(result)

        if is_single_image:
            return results[0]

        return results



# pyright: reportShadowedImports=false
import math
from typing import *

import numpy as np
from mousse import Dataclass

from ...common import Detection, FaceDetection, Point, clip, nms, sigmoid
from ...runner import Runner, RunnerConfig

__all__ = ["Yolov5FacePostProcessRunner", "Yolov5FacePostProcessRunnerConfig"]

Size = Tuple[int, int]
Dims = Tuple[int, int, int, int, int]


class Yolov5FaceGrid(Dataclass):
    width: int
    height: int
    x: int
    y: int
    idx: int


class Yolov5FacePostProcessRunnerConfig(RunnerConfig):
    type: str = "Yolov5FacePostProcessRunner"

    outputs: Dict[str, Dims] = {}
    anchors: Dict[str, List[Size]] = []

    conf_threshold: float = 0.5
    nms_threshold: float = 0.3

    top_k: int = 1000

    height: int = -1
    width: int = -1
    batch_size: int = -1


class YoloxPostProcessInputData(Dataclass):
    images: np.ndarray
    output: np.ndarray


class Yolov5FacePostProcessRunner(Runner):
    config_type = Yolov5FacePostProcessRunnerConfig
    outputs = ["detections"]

    def prepare(self) -> "Yolov5FacePostProcessRunner":
        config: Yolov5FacePostProcessRunnerConfig = self.config

        self.grids: Dict[str, List[Yolov5FaceGrid]] = {}
        if config.width > 0 and config.height > 0:
            self.generate_grids()

        return self

    def generate_grids(self):
        config: Yolov5FacePostProcessRunnerConfig = self.config
        self.grids.clear()

        for key, dims in config.outputs.items():
            anchors = config.anchors[key]

            _, num_anchors_per_layer, num_grid_h, num_grid_w, _ = dims
            assert num_anchors_per_layer == len(
                anchors
            ), f"Mismatch num anchors per layer for {key}"

            grids = []
            for idx, (width, height) in enumerate(anchors):
                for y in range(num_grid_h):
                    for x in range(num_grid_w):
                        grids.append(
                            Yolov5FaceGrid(
                                x=x, y=y, width=width, height=height, idx=idx
                            )
                        )

            self.grids[key] = grids

    def generate_proposals(self, key: str, output: np.ndarray) -> List[Detection]:
        config: Yolov5FacePostProcessRunnerConfig = self.config

        proposals = []
        grids: List[Yolov5FaceGrid] = self.grids[key]
        _, num_grid_h, num_grid_w, num_features = output.shape

        for grid in grids:
            tensor = output[grid.idx, grid.y, grid.x, ...]

            stride_h = config.height / num_grid_h
            stride_w = config.width / num_grid_w

            obj_score = sigmoid(tensor[4])
            if obj_score < config.conf_threshold:
                continue

            cls_score = sigmoid(tensor[-1])
            if cls_score < config.conf_threshold:
                continue

            dx, dy, dw, dh = sigmoid(tensor[:4])

            x_center = (2 * dx - 0.5 + grid.x) * stride_w
            y_center = (2 * dy - 0.5 + grid.y) * stride_h
            width = math.pow(2 * dw, 2) * grid.width
            height = math.pow(2 * dh, 2) * grid.height

            proposal = FaceDetection(
                left=x_center - width / 2,
                top=y_center - height / 2,
                right=x_center + width / 2,
                bottom=y_center + height / 2,
                prob=cls_score,
            )

            num_keypoints = (num_features - 6) // 2
            for i in range(0, 2 * num_keypoints, 2):
                dx = float(tensor[i + 5])
                dy = float(tensor[i + 6])

                proposal.keypoints.append(
                    Point(
                        x=dx * grid.width + grid.x * stride_w,
                        y=dy * grid.height + grid.y * stride_h,
                    )
                )
            proposals.append(proposal)

        return proposals

    def run(self, **outputs: Dict[str, np.ndarray]) -> List[List[Detection]]:
        config: Yolov5FacePostProcessRunnerConfig = self.config

        min_batch_size = min(tensor.shape[0] for tensor in outputs.values())
        max_batch_size = max(tensor.shape[0] for tensor in outputs.values())
        assert min_batch_size == max_batch_size, f"Batch size mismatch"

        keys = list(outputs.keys())
        batch_size = min_batch_size

        outputs = [{key: outputs[key][i] for key in outputs} for i in range(batch_size)]

        batch_proposals = []
        for sample_id in range(batch_size):
            sample_proposals = []
            for key in keys:
                proposals = self.generate_proposals(key, outputs[sample_id][key])
                sample_proposals.extend(proposals)

            sample_proposals.sort(key=lambda proposal: proposal.prob, reverse=True)
            batch_proposals.append(sample_proposals[: config.top_k])

        batch_objects = []
        for proposals in batch_proposals:
            proposals = nms(proposals, config.nms_threshold)

            objects = []
            for proposal in proposals:
                proposal.left = clip(proposal.left, 0, config.width)
                proposal.top = clip(proposal.top, 0, config.height)
                proposal.right = clip(proposal.right, 0, config.width)
                proposal.bottom = clip(proposal.bottom, 0, config.height)

                if proposal.right <= proposal.left or proposal.bottom <= proposal.top:
                    continue

                proposal.left /= config.width
                proposal.top /= config.height
                proposal.right /= config.width
                proposal.bottom /= config.height

                for point in proposal.keypoints:
                    point.x /= config.width
                    point.y /= config.height

                objects.append(proposal)

            batch_objects.append(objects)

        return batch_objects



# pyright: reportShadowedImports=false
from typing import *

from ...runner import RunnerConfig
from ..preproc import ImagePreProcessRunner, ImagePreProcessRunnerConfig

__all__ = ["YoloXPreProcessRunner", "YoloXPreProcessRunnerConfig"]


class YoloXPreProcessRunnerConfig(ImagePreProcessRunnerConfig, dynamic=True):
    type: str = "YoloXPreProcessRunner"


class YoloXPreProcessRunner(ImagePreProcessRunner):
    config_type: Type[RunnerConfig] = YoloXPreProcessRunnerConfig

    def prepare(self) -> "YoloXPreProcessRunner":
        return self



# pyright: reportShadowedImports=false
from typing import *

import numpy as np

from ...common import Detection, Image, Images, InputFormat
from ...runner import Runner, RunnerConfig
from ..onnx import OnnxNodeInfo, OnnxRunner, OnnxRunnerConfig
from .postproc import ICLSPostProcessRunner, ICLSPostProcessRunnerConfig
from .preproc import ICLSPreProcessRunner, ICLSPreProcessRunnerConfig

__all__ = ["ICLSModelRunner", "ICLSModelRunnerConfig"]


class ICLSModelRunnerConfig(
    ICLSPreProcessRunnerConfig,
    OnnxRunnerConfig,
    ICLSPostProcessRunnerConfig,
    dynamic=True,
):
    type: str = "ICLSModelRunner"
    input_node: str = "images"
    output_node: str = "output"
    batch_size: int = -1


class ICLSModelRunner(Runner):
    config_type: Type[RunnerConfig] = ICLSModelRunnerConfig

    def prepare(self) -> "ICLSModelRunner":
        config: ICLSModelRunner = self.config

        self.onnx_runner = OnnxRunner(config)
        self.onnx_runner.prepare()

        self.input_info: OnnxNodeInfo = self.onnx_runner.metadata[config.input_node]

        if config.input_format == InputFormat.hwc:
            (
                config.batch_size,
                config.height,
                config.width,
                _,
            ) = self.input_info.dims

        if config.input_format == InputFormat.chw:
            (
                config.batch_size,
                _,
                config.height,
                config.width,
            ) = self.input_info.dims

        self.preproc_runner = ICLSPreProcessRunner(config)
        self.preproc_runner.prepare()

        self.postproc_runner = ICLSPostProcessRunner(config)
        self.postproc_runner.prepare()

        return self

    def run(
        self, images: Union[Image, Images], **kwargs
    ) -> Union[List[Detection], List[List[Detection]]]:
        config: ICLSModelRunnerConfig = self.config

        is_single_image = type(images) is not list
        if is_single_image:
            images = [images]

        images = [
            Image(mat=image) if isinstance(image, np.ndarray) else image
            for image in images
        ]

        preproc_images = []
        for image in images:
            preproc_image = self.preproc_runner.run(image)
            preproc_images.append(preproc_image.mat)

        results = []
        for i in range(0, len(preproc_images), config.batch_size):
            batch = preproc_images[i : i + config.batch_size]
            batch_size = len(batch)
            if batch_size < config.batch_size:
                batch += [
                    np.zeros(self.input_info.dims[1:], dtype=self.input_info.dtype)
                    for _ in range(config.batch_size - batch_size)
                ]

            batch = np.stack(batch)
            output = self.onnx_runner.run(**{config.input_node: batch})[
                config.output_node
            ]

            batch = batch[:batch_size, ...]
            output = output[:batch_size, ...]

            classifications = self.postproc_runner.run(output=output)
            results.extend(classifications)

        if is_single_image:
            return results[0]

        return results



# pyright: reportShadowedImports=false
from collections import deque
from typing import *

import numpy as np

from ...common import Classification, softmax
from ...runner import Runner, RunnerConfig

__all__ = ["TSMPostProcessRunner", "TSMPostProcessRunnerConfig"]


class TSMPostProcessRunnerConfig(RunnerConfig, dynamic=True):
    type: str = "TSMPostProcessRunnerConfig"
    classes: List[str]
    weights: List[float] = []

    max_segments: int = 30
    segment_size: int = 8
    overlap_size: int = 4
    step: int = 8

    @property
    def num_classes(self):
        return len(self.classes)


class TSMPostProcessRunner(Runner):
    config_type: Type[RunnerConfig] = TSMPostProcessRunnerConfig
    outputs: List[str] = ["classifications"]

    def prepare(self) -> "TSMPostProcessRunner":
        config: TSMPostProcessRunnerConfig = self.config
        assert (
            config.overlap_size < config.segment_size
        ), f"Segment size must be greater than overlap size"

        if len(config.weights) == 0:
            config.weights = [1] * config.num_classes

        assert len(config.weights) == config.num_classes, f"Invalid weight config"

        self.cls_scores = deque(maxlen=config.max_segments * config.segment_size)
        return self

    def reset(self):
        self.cls_scores.clear()

    def run(self, output: np.ndarray, **kwargs) -> List[Classification]:
        config: TSMPostProcessRunnerConfig = self.config
        batch_size, *_ = output.shape

        scores = []
        num_supporters = []

        for i in range(batch_size):
            self.cls_scores.append(output[i, ...])

            cls_scores = np.asarray(self.cls_scores)

            num_frames, *_ = cls_scores.shape
            index = num_frames - 1
            indices = []
            segment = []
            while True:
                segment.append(index)
                if len(segment) == config.segment_size:
                    indices.extend(segment)
                    index = segment[-config.overlap_size]
                    segment.clear()
                    continue

                index -= config.step
                if index < 0:
                    break

            segment_size = config.segment_size
            if not indices:
                segment_size = len(segment)
                indices.extend(segment)

            indices.reverse()

            cls_scores = cls_scores.take(indices, axis=0)
            cls_scores = cls_scores.reshape((-1, segment_size, config.num_classes))
            num_segments, *_ = cls_scores.shape
            num_supporters.append(num_segments * segment_size)

            avg_scores = np.mean(cls_scores, axis=1)
            scores.append(softmax(avg_scores))

        scores = np.asarray(scores)
        scores = np.mean(scores, axis=1)

        _, num_classes = scores.shape
        assert num_classes == len(config.classes), f"Missing classes"

        # labels = np.argmax(scores, axis=-1)
        classifications = []

        for i, (score, num_support) in enumerate(zip(scores, num_supporters)):
            classifications.append(
                Classification(scores=score, support=num_support, labels=config.classes)
            )

        return classifications



# pyright: reportShadowedImports=false
import inspect
from abc import abstractproperty
from pathlib import Path
from typing import *

from mousse import AutoRegistry, Dataclass, Registry, asclass
from mousse.types.parser import asdict
from typer import Argument

from .cli import cli

__all__ = ["runner_registry", "Runner", "RunnerConfig", "query"]

runner_registry = Registry.get("runner")


class RunnerConfig(Dataclass, dynamic=True):
    type: str
    # tag: str = None
    # inputs: Dict[str, str] = {}
    # outputs: Dict[str, str] = {}
    # metadata: Dict[str, str] = {}


class Runner(metaclass=AutoRegistry, registry="runner"):
    def __init__(self, config: RunnerConfig) -> None:
        if type(config) is not self.config_type:
            config = asclass(self.config_type, asdict(config))

        self.config = config
        self.metadata = {}

    @property
    def config_type(self) -> Type[RunnerConfig]:
        return RunnerConfig

    def prepare(self) -> "Runner":
        pass

    @abstractproperty
    def outputs(self) -> List[str]:
        pass

    def run(self, *args, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError()

    def reset(self):
        pass

    def __call__(self, **inputs: Dict[str, Any]) -> Dict[str, Any]:
        return self.run(**inputs)

    @classmethod
    def create(cls, config: RunnerConfig = None, path: Union[str, Path] = None):
        assert (config is not None) ^ (path is not None), f"Invalid argument"

        if path is not None:
            config = asclass(RunnerConfig, path=path)

        runner_cls = runner_registry[config.type]
        return runner_cls(config)


@cli.command()
def query(runner: str = Argument(..., help="Runner")):
    runner_cls = runner_registry[runner]
    signature = inspect.signature(runner_cls.run)

    print("Inputs:")
    for name, info in signature.parameters.items():
        if name == "self":
            continue

        if info.kind == info.VAR_POSITIONAL:
            name = f"*{name}"

        if info.kind == info.VAR_KEYWORD:
            name = f"**{name}"

        annotation = info.annotation if info.annotation != inspect._empty else Any
        default = info.default if info.default != inspect._empty else ""

        annotation = str(annotation)

        summary = f"\t{name: <16}: {annotation: <8} [{default}]"
        print(summary)

    print("Outputs")
    print(f"\t{signature.return_annotation}")


@cli.command()
def ls():
    runners_cls = list(runner_registry.keys())
    runners_cls.sort()
    for key in runners_cls:
        if key != "Runner":
            print(key)



# pyright: reportShadowedImports=false
import inspect
from functools import lru_cache, wraps
from inspect import Signature
from types import FunctionType
from typing import *

from typer import Typer
from typer.models import ArgumentInfo, OptionInfo

__all__ = ["cli", "Command", "CLI"]


class Command:
    pass


class CLI(Typer):
    @wraps(Typer.command)
    def command(self, *args, callback: Callable = None, **kwargs):
        base = super()

        def decorator(func: Callable):
            __func = func
            if callback is not None:

                @wraps(func)
                def __func(*args, **kwargs):
                    return callback(func(*args, **kwargs))

            base.command(*args, **kwargs)(__func)
            func = FunctionType(
                func.__code__,
                func.__globals__,
                func.__name__,
                func.__defaults__,
                func.__closure__,
            )
            signature = inspect.signature(func)
            parameters = list(signature.parameters.values())
            defaults = []

            for i, parameter in enumerate(parameters):
                if type(parameter.default) == ArgumentInfo:
                    parameter = parameter.replace(default=inspect._empty)
                    parameters[i] = parameter

                if type(parameter.default) == OptionInfo:
                    parameter = parameter.replace(default=parameter.default.default)
                    parameters[i] = parameter

                if parameter.default != inspect._empty:
                    defaults.append(parameter.default)

            setattr(
                func,
                "__signature__",
                Signature(
                    parameters=parameters, return_annotation=signature.return_annotation
                ),
            )
            func.__defaults__ = tuple(defaults)

            class ConcreteCommand(Command):
                @wraps(func)
                def __call__(self, *args: Any, **kwds: Any):
                    return func(*args, **kwds)

            return ConcreteCommand()

        return decorator

    @lru_cache(maxsize=None)
    def sub_cli(self, name: str):
        sub_app = CLI(pretty_exceptions_show_locals=False)
        self.add_typer(sub_app, name=name)
        return sub_app


cli = CLI(pretty_exceptions_show_locals=False)


@cli.command()
def about():
    import random

    import cowsay

    char = random.choice(cowsay.char_names)
    print(cowsay.get_output_string(char, "This is a module for running pipeline"))



import numpy as np
import scipy.linalg

__all__ = ["KalmanFilter"]


class KalmanFilter:
    """
    A simple Kalman filter for tracking bounding boxes in image space.

    The 8-dimensional state space

        x, y, a, h, vx, vy, va, vh

    contains the bounding box center position (x, y), aspect ratio a, height h,
    and their respective velocities.

    Object motion follows a constant velocity model. The bounding box location
    (x, y, a, h) is taken as direct observation of the state space (linear
    observation model).

    """

    def __init__(self, pos_weight: float = 1.0 / 20, vel_weight: float = 1.0 / 160):
        ndim, dt = 4, 1.0

        # Create Kalman filter model matrices.
        self._motion_mat: np.ndarray = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat: np.ndarray = np.eye(ndim, 2 * ndim)

        # Motion and observation uncertainty are chosen relative to the current
        # state estimate. These weights control the amount of uncertainty in
        # the model. This is a bit hacky.
        self._pos_weight: float = pos_weight
        self._vel_weight: float = vel_weight

        self.mean: np.ndarray = None
        self.covariance: np.ndarray = None

    def initiate(self, measurement: np.ndarray):
        """Create track from unassociated measurement.

        Parameters
        ----------
        measurement : ndarray
            Bounding box coordinates (x, y, a, h) with center position (x, y),
            aspect ratio a, and height h.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector (8 dimensional) and covariance matrix (8x8
            dimensional) of the new track. Unobserved velocities are initialized
            to 0 mean.

        """
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        self.mean = np.r_[mean_pos, mean_vel]

        std = [
            2 * self._pos_weight * measurement[3],
            2 * self._pos_weight * measurement[3],
            1e-2,
            2 * self._pos_weight * measurement[3],
            10 * self._vel_weight * measurement[3],
            10 * self._vel_weight * measurement[3],
            1e-5,
            10 * self._vel_weight * measurement[3],
        ]
        self.covariance = np.diag(np.square(std))
        
        return self.mean, self.covariance

    def predict(self):
        """Run Kalman filter prediction step.

        Parameters
        ----------
        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector and covariance matrix of the predicted
            state. Unobserved velocities are initialized to 0 mean.

        """
        std_pos = [
            self._pos_weight * self.mean[3],
            self._pos_weight * self.mean[3],
            1e-2,
            self._pos_weight * self.mean[3],
        ]
        std_vel = [
            self._vel_weight * self.mean[3],
            self._vel_weight * self.mean[3],
            1e-5,
            self._vel_weight * self.mean[3],
        ]
        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        self.mean = np.dot(self.mean, self._motion_mat.T)
        self.covariance = (
            np.linalg.multi_dot((self._motion_mat, self.covariance, self._motion_mat.T))
            + motion_cov
        )

        return self.mean, self.covariance

    def project(self, mean: np.ndarray, covariance: np.ndarray):
        """Project state distribution to measurement space.

        Parameters
        ----------
        mean : ndarray
            The state's mean vector (8 dimensional array).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).

        Returns
        -------
        (ndarray, ndarray)
            Returns the projected mean and covariance matrix of the given state
            estimate.

        """
        std = [
            self._pos_weight * mean[3],
            self._pos_weight * mean[3],
            1e-1,
            self._pos_weight * mean[3],
        ]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot(
            (self._update_mat, covariance, self._update_mat.T)
        )
        return mean, covariance + innovation_cov

    def update(self, measurement: np.ndarray):
        """Run Kalman filter correction step.

        Parameters
        ----------
        mean : ndarray
            The predicted state's mean vector (8 dimensional).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).
        measurement : ndarray
            The 4 dimensional measurement vector (x, y, a, h), where (x, y)
            is the center position, a the aspect ratio, and h the height of the
            bounding box.

        Returns
        -------
        (ndarray, ndarray)
            Returns the measurement-corrected state distribution.

        """
        projected_mean, projected_cov = self.project(self.mean, self.covariance)

        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False
        )
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower),
            np.dot(self.covariance, self._update_mat.T).T,
            check_finite=False,
        ).T
        innovation = measurement - projected_mean

        self.mean = self.mean + np.dot(innovation, kalman_gain.T)
        self.covariance = self.covariance - np.linalg.multi_dot(
            (kalman_gain, projected_cov, kalman_gain.T)
        )
        return self.mean, self.covariance

    # def gating_distance(
    #     self,
    #     mean: np.ndarray,
    #     covariance: np.ndarray,
    #     measurements: np.ndarray,
    #     only_position: bool = False,
    #     metric: str = "maha",
    # ):
    #     """Compute gating distance between state distribution and measurements.
    #     A suitable distance threshold can be obtained from `chi2inv95`. If
    #     `only_position` is False, the chi-square distribution has 4 degrees of
    #     freedom, otherwise 2.
    #     Parameters
    #     ----------
    #     mean : ndarray
    #         Mean vector over the state distribution (8 dimensional).
    #     covariance : ndarray
    #         Covariance of the state distribution (8x8 dimensional).
    #     measurements : ndarray
    #         An Nx4 dimensional matrix of N measurements, each in
    #         format (x, y, a, h) where (x, y) is the bounding box center
    #         position, a the aspect ratio, and h the height.
    #     only_position : Optional[bool]
    #         If True, distance computation is done with respect to the bounding
    #         box center position only.
    #     Returns
    #     -------
    #     ndarray
    #         Returns an array of length N, where the i-th element contains the
    #         squared Mahalanobis distance between (mean, covariance) and
    #         `measurements[i]`.
    #     """
    #     mean, covariance = self.project(mean, covariance)
    #     if only_position:
    #         mean, covariance = mean[:2], covariance[:2, :2]
    #         measurements = measurements[:, :2]

    #     d = measurements - mean
    #     if metric == "gaussian":
    #         return np.sum(d * d, axis=1)
    #     elif metric == "maha":
    #         cholesky_factor = np.linalg.cholesky(covariance)
    #         z = scipy.linalg.solve_triangular(
    #             cholesky_factor, d.T, lower=True, check_finite=False, overwrite_b=True
    #         )
    #         squared_maha = np.sum(z * z, axis=0)
    #         return squared_maha
    #     else:
    #         raise ValueError("invalid distance metric")



# pyright: reportShadowedImports=false
from typing import *

import numpy as np

from ...common import Detection, clip, nms
from ...runner import Runner, RunnerConfig

__all__ = ["FairMOTPostProcessRunner", "FairMOTPostProcessRunnerConfig"]


class FairMOTPostProcessRunnerConfig(RunnerConfig):
    type: str = "FairMOTPostProcessRunner"
    hm_layer: str = "hm"
    wh_layer: str = "wh"
    reg_layer: str = "reg"

    conf_threshold: float = 0.5
    nms_threshold: float = 0.3

    top_k: int = 200

    height: int = -1
    width: int = -1
    batch_size: int = -1


class FairMOTPostProcessRunner(Runner):
    config_type = FairMOTPostProcessRunnerConfig
    outputs = ["detections"]

    def prepare(self) -> "FairMOTPostProcessRunner":
        config: FairMOTPostProcessRunnerConfig = self.config
        self.mapping = {
            config.hm_layer: "hm",
            config.wh_layer: "wh",
            config.reg_layer: "reg",
        }
        return self

    def generate_proposals(
        self, hm: np.ndarray, wh: np.ndarray, reg: np.ndarray
    ) -> List[Detection]:
        config: FairMOTPostProcessRunnerConfig = self.config

        _, output_height, output_width = hm.shape
        scores = []
        for row in range(output_height):
            for col in range(output_width):
                score = float(hm[0, row, col])
                if score > config.conf_threshold:
                    scores.append((score, row, col))

        scores.sort(key=lambda val: val[0], reverse=True)
        scores = scores[: config.top_k]

        proposals = []
        for score, row, col in scores:
            cx_offset, cy_offset = reg[row, col, :]
            left, top, right, bottom = wh[row, col, :]

            cx = col + cx_offset
            cy = row + cy_offset

            detection = Detection(
                prob=score,
                left=clip((cx - left) / output_width, 0, 1),
                top=clip((cy - top) / output_height, 0, 1),
                right=clip((cx + right) / output_width, 0, 1),
                bottom=clip((cy + bottom) / output_height, 0, 1),
            )
            # print(detection)

            proposals.append(detection)

        return proposals

    def run(self, **outputs: Dict[str, np.ndarray]) -> List[List[Detection]]:
        config: FairMOTPostProcessRunnerConfig = self.config

        outputs = {
            self.mapping[key]: val
            for key, val in outputs.items()
            if key in self.mapping
        }
        assert len(outputs) == len(self.mapping), "Missing keys"

        min_batch_size = min(tensor.shape[0] for tensor in outputs.values())
        max_batch_size = max(tensor.shape[0] for tensor in outputs.values())
        assert min_batch_size == max_batch_size, f"Batch size mismatch"

        batch_size = min_batch_size
        outputs = [{key: outputs[key][i] for key in outputs} for i in range(batch_size)]

        batch_proposals = []
        for output in outputs:
            batch_proposals.append(self.generate_proposals(**output))

        batch_objects = []
        for proposals in batch_proposals:
            proposals = nms(proposals, config.nms_threshold)

            objects = []
            for proposal in proposals:
                if proposal.right <= proposal.left or proposal.bottom <= proposal.top:
                    continue

                objects.append(proposal)

            batch_objects.append(objects)

        return batch_objects



# pyright: reportShadowedImports=false
from typing import *

from ...runner import RunnerConfig
from ..preproc import ImagePreProcessRunner, ImagePreProcessRunnerConfig

__all__ = ["FairMOTPreProcessRunner", "FairMOTPreProcessRunnerConfig"]


class FairMOTPreProcessRunnerConfig(ImagePreProcessRunnerConfig, dynamic=True):
    type: str = "FairMOTPreProcessRunner"


class FairMOTPreProcessRunner(ImagePreProcessRunner):
    config_type: Type[RunnerConfig] = FairMOTPreProcessRunnerConfig

    def prepare(self) -> "FairMOTPreProcessRunner":
        return self



# pyright: reportShadowedImports=false
import math
from typing import *

import numpy as np

from ...common import Detection, Image, Images, InputFormat, clip
from ...runner import Runner, RunnerConfig
from ..onnx import OnnxNodeType, OnnxRunner, OnnxRunnerConfig
from .postproc import FairMOTPostProcessRunner, FairMOTPostProcessRunnerConfig
from .preproc import FairMOTPreProcessRunner, FairMOTPreProcessRunnerConfig

__all__ = ["FairMOTModelRunner", "FairMOTModelRunnerConfig"]


class FairMOTModelRunnerConfig(
    FairMOTPreProcessRunnerConfig,
    OnnxRunnerConfig,
    FairMOTPostProcessRunnerConfig,
    dynamic=True,
):
    type: str = "FairMOTModelRunner"


class FairMOTModelRunner(Runner):
    config_type: Type[RunnerConfig] = FairMOTModelRunnerConfig
    outputs: List[str] = ["detections"]

    def prepare(self) -> "FairMOTModelRunner":
        config: FairMOTModelRunnerConfig = self.config

        self.onnx_runner = OnnxRunner(config)
        self.onnx_runner.prepare()

        for _, node_info in self.onnx_runner.metadata.items():
            if node_info.ntype == OnnxNodeType.input:
                self.input_info = node_info
                if config.input_format == InputFormat.hwc:
                    (
                        config.batch_size,
                        config.height,
                        config.width,
                        _,
                    ) = self.input_info.dims

                if config.input_format == InputFormat.chw:
                    (
                        config.batch_size,
                        _,
                        config.height,
                        config.width,
                    ) = self.input_info.dims

        self.preproc_runner = FairMOTPreProcessRunner(config)
        self.preproc_runner.prepare()

        self.postproc_runner = FairMOTPostProcessRunner(config)
        self.postproc_runner.prepare()

        return self

    def run(
        self, images: Union[Image, Images], **kwargs
    ) -> Union[List[Detection], List[List[Detection]]]:
        config: FairMOTModelRunnerConfig = self.config

        is_single_image = type(images) is not list
        if is_single_image:
            images = [images]

        images = [
            Image(mat=image) if isinstance(image, np.ndarray) else image
            for image in images
        ]

        preproc_images = []
        for image in images:
            preproc_image = self.preproc_runner.run(image)
            preproc_images.append(preproc_image.mat)

        candidates = []
        for i in range(0, len(preproc_images), config.batch_size):
            batch = preproc_images[i : i + config.batch_size]
            batch_size = len(batch)
            if batch_size < config.batch_size:
                batch += [
                    np.zeros(self.input_info.dims[1:], dtype=self.input_info.dtype)
                    for _ in range(config.batch_size - batch_size)
                ]

            batch = np.stack(batch)
            outputs = self.onnx_runner.run(**{self.input_info.name: batch})

            batch = batch[:batch_size, ...]
            outputs = {key: val[:batch_size, ...] for key, val in outputs.items()}

            detections = self.postproc_runner.run(**outputs)
            candidates.extend(detections)

        results = []
        for candidate, image in zip(candidates, images):
            height, width, _ = image.mat.shape
            top, bottom, left, right = image.padding

            result = []
            for detection in candidate:
                detection.left = clip(detection.left - left, 0, 1)
                detection.top = clip(detection.top - top, 0, 1)
                detection.right = clip(detection.right - right, 0, 1)
                detection.bottom = clip(detection.bottom - bottom, 0, 1)

                if (
                    detection.right <= detection.left
                    or detection.bottom <= detection.top
                ):
                    continue

                detection.left = math.floor(detection.left * width)
                detection.top = math.floor(detection.top * height)
                detection.right = math.floor(detection.right * width)
                detection.bottom = math.floor(detection.bottom * height)

                result.append(detection)

        results.append(result)

        if is_single_image:
            return results[0]

        return results



# pyright: reportShadowedImports=false
from typing import *

from mousse import Dataclass
from .bbox import BBox, BBoxType

__all__ = ["Point", "Classification", "Detection", "FaceDetection"]


class Point(Dataclass, dynamic=True):
    x: float
    y: float


class Classification(Dataclass, dynamic=True):
    scores: List[float]
    labels: List[str]

    def __getitem__(self, label: str):
        if label not in self.labels:
            raise IndexError(label)

        return self.scores[self.labels.index(label)]


class Detection(Dataclass, dynamic=True):
    left: float
    top: float
    right: float
    bottom: float
    prob: float
    label: int

    track_id: int = -1

    @property
    def width(self) -> float:
        return self.right - self.left

    @property
    def height(self) -> float:
        return self.bottom - self.top

    def to(self, btype: BBoxType) -> BBox:
        bbox = BBox(
            rect=(self.left, self.top, self.right, self.bottom),
            type=BBoxType.ltrb,
            score=self.prob,
        )

        if btype == BBoxType.ltrb:
            return bbox

        return bbox.to(btype)


class FaceDetection(Detection, dynamic=True):
    keypoints: List[Point] = []



# pyright: reportShadowedImports=false
from typing import *

import numpy as np
from scipy.optimize import linear_sum_assignment

from ...runner import Runner, RunnerConfig
from ...common import Detection, BBox, iou
from .track import Track, TrackState


__all__ = ["ByteTrackRunner", "ByteTrackRunnerConfig"]


def area_similarity(target_area: float, ref_bbox: BBox) -> float:
    similarity = 1 - abs(ref_bbox.area - target_area) / target_area
    return max(0, similarity)


def linear_assignment(cost_matrix: np.ndarray, threshold: float):
    if cost_matrix.size == 0:
        return (
            np.empty((0, 2), dtype=int),
            tuple(range(cost_matrix.shape[0])),
            tuple(range(cost_matrix.shape[1])),
        )
    candidate_rows, candidate_cols = linear_sum_assignment(cost_matrix)

    matched_rows = np.zeros(cost_matrix.shape[0])
    matches_cols = np.zeros(cost_matrix.shape[1])

    matches = []
    for row, col in zip(candidate_rows, candidate_cols):
        cost = cost_matrix[row, col]
        if cost < threshold:
            matches.append((row, col))
            matched_rows[row] = 1
            matches_cols[col] = 1

    matches = np.asarray(matches)
    unmatched_a = np.where(matched_rows == 0)[0]
    unmatched_b = np.where(matches_cols == 0)[0]

    return matches, unmatched_a, unmatched_b


def ious(a_bboxes: List[BBox], b_bboxes: List[BBox]) -> np.ndarray:
    """
    Compute cost based on IoU
    :rtype ious np.ndarray
    """
    _ious = np.zeros((len(a_bboxes), len(b_bboxes)), dtype=float)
    if _ious.size == 0:
        return _ious

    for i, a_bbox in enumerate(a_bboxes):
        for j, b_bbox in enumerate(b_bboxes):
            _ious[i, j] = iou(a_bbox, b_bbox)

    return _ious


def localization_cost(
    tracks: Sequence[Track],
    detections: Sequence[Detection],
    iou_weight: float = 0.6,
    diff_weight: float = 0.4,
) -> np.ndarray:
    detections = [detection.to("ltrb") for detection in detections]

    curr_ious = ious([track.pred_bbox for track in tracks], detections)

    area_similarity_matrix = np.zeros(curr_ious.shape)
    inds = np.nonzero(curr_ious > 0)

    tracks_mean_area = [track.mean_area for track in tracks]
    for i, j in zip(inds[0], inds[1]):
        area_similarity_matrix[i][j] = area_similarity(tracks_mean_area[i], detections[j])

    cost_matrix = 1 - (curr_ious * iou_weight + area_similarity_matrix * diff_weight)
    return cost_matrix


def fuse_score(
    cost_matrix: np.ndarray,
    detections: Sequence[Detection],
    tracks: Sequence[Track],
    score_weight: float = 0.3,
    use_label: bool = False,
) -> np.ndarray:
    if cost_matrix.size == 0:
        return cost_matrix

    det_scores = np.array([det.prob for det in detections])
    det_scores = np.expand_dims(det_scores, axis=0).repeat(cost_matrix.shape[0], axis=0)

    mean_scores = np.array([[track.mean_score] for track in tracks])
    assert mean_scores.shape[0] == cost_matrix.shape[0]
    mean_scores = np.repeat(mean_scores, det_scores.shape[1], axis=1)
    
    det_scores = score_weight * mean_scores + (1 - score_weight) * det_scores

    similariry = (1 - cost_matrix) * det_scores
    cost = 1 - similariry

    if use_label:
        for i, track in enumerate(tracks):
            for j, detection in enumerate(detections):
                if track.detection.label != detection.label:
                    cost[i, j] = 1

    return cost


class ByteTrackRunnerConfig(RunnerConfig, dynamic=True):
    type: str = "ByteTrackRunner"

    match_threshold: float = 0.85
    unconfirmed_threshold: float = 0.7
    link_threshold: float = 0.85
    max_time_lost: int = 30
    max_track_id: int = 10000

    max_history: int = 5
    min_history: int = 1

    use_label: bool = False


class ByteTrackRunner(Runner):
    config_type: Type[RunnerConfig] = ByteTrackRunnerConfig
    outputs: List[str] = ["tracks"]

    def prepare(self) -> "ByteTrackRunner":
        self.tracks: List[Track] = []
        self.frame_id = 0
        self.track_id = 0

    def reset(self):
        self.tracks.clear()
        self.frame_id = 0
        self.track_id = 0

    def next_id(self):
        self.track_id = (self.track_id + 1) % self.config.max_track_id
        return self.track_id

    def link_tracks_and_detections(
        self,
        tracks: Sequence[Track],
        detections: Sequence[Detection],
        threshold: float = 0.85,
    ):
        det_ltrbs = [det.to("ltrb") for det in detections]
        track_ltrbs = [
            track.last_bbox.to("ltrb")
            if track.state != TrackState.Tracked
            else track.pred_bbox.to("ltrb")
            for track in tracks
        ]
        dists = 1 - ious(track_ltrbs, det_ltrbs)
        matching_pairs = {}
        confused_tracks = []
        for i, _ in enumerate(dists):
            candidate_ids = [j for j, dist in enumerate(dists[i]) if dist < 1]
            if len(candidate_ids) != 0:
                for j in candidate_ids:
                    if j not in matching_pairs:
                        matching_pairs[j] = [i]
                    else:
                        matching_pairs[j].append(i)
                if len(candidate_ids) > 1:
                    confused_tracks.append(i)

        linked_tracks, linked_dets = [], []
        for det_idx, matching_tracks in matching_pairs.items():
            if len(matching_tracks) == 1 and matching_tracks[0] not in confused_tracks:
                matching_track_id = matching_tracks[0]
                _area_similarity = area_similarity(
                    tracks[matching_track_id].mean_area, det_ltrbs[det_idx]
                )
                if _area_similarity > threshold:
                    linked_dets.append(det_idx)
                    linked_tracks.append(matching_track_id)

        return linked_tracks, linked_dets

    def get_comfirmed_tracks(self):
        tracks_pool: Sequence[Track] = []
        for track in self.tracks:
            if track.state != TrackState.New:
                tracks_pool.append(track)

        return tracks_pool

    def get_unmatched_detections(self, detections: Sequence[Detection]):
        return [detection for detection in detections if not detection.matched]

    def get_unmatched_tracks(self, state: TrackState = None):
        tracks_pool = self.get_comfirmed_tracks()
        return [
            track
            for track in tracks_pool
            if track.frame_id != self.frame_id
            and (state is None or track.state == state)
        ]

    def associate_confirmed_tracks(self, detections: Sequence[Detection]):
        tracks_pool = self.get_comfirmed_tracks()

        for track in tracks_pool:
            track.predict()

        matches, *_ = linear_assignment(
            fuse_score(
                cost_matrix=localization_cost(tracks_pool, detections),
                detections=detections,
                tracks=tracks_pool,
                use_label=self.config.use_label,
            ),
            threshold=self.config.match_threshold,
        )

        for itrack, idet in matches:
            track = tracks_pool[itrack]
            detection = detections[idet]
            detection.matched = True
            track.update(detection, self.frame_id)

        u_detections = self.get_unmatched_detections(detections)
        u_tracks = self.get_unmatched_tracks(TrackState.Tracked)

        linked_tracks, linked_dets = self.link_tracks_and_detections(
            u_tracks, u_detections, self.config.link_threshold
        )
        for track_idx, det_idx in zip(linked_tracks, linked_dets):
            track = u_tracks[track_idx]
            detection = u_detections[det_idx]
            detection.matched = True
            track.update(detection, self.frame_id)

    def get_unconfirmed_trasks(self):
        u_tracks: Sequence[Track] = []
        for track in self.tracks:
            if track.state == TrackState.New:
                u_tracks.append(track)

        return u_tracks

    def associate_unconfirmed_tracks(self, detections: Sequence[Detection]):
        u_tracks = self.get_unconfirmed_trasks()
        u_detections = self.get_unmatched_detections(detections)
        
        dists = fuse_score(
            1 - ious([
                track.last_bbox for track in u_tracks
            ], u_detections),
            detections=u_detections,
            tracks=u_tracks,
            use_label=self.config.use_label,
        )

        matches, _, _ = linear_assignment(
            dists, threshold=self.config.unconfirmed_threshold
        )
        for itrack, idet in matches:
            detection = u_detections[idet]
            track = u_tracks[itrack]
            detection.matched = True
            track.update(detection, self.frame_id)

        u_tracks = self.get_unconfirmed_trasks()
        u_detections = self.get_unmatched_detections(detections)
        linked_tracks, linked_dets = self.link_tracks_and_detections(
            u_tracks, u_detections, threshold=self.config.link_threshold
        )
        for track_idx, det_idx in zip(linked_tracks, linked_dets):
            track = u_tracks[track_idx]
            detection = u_detections[det_idx]
            detection.matched = True
            track.update(detection, self.frame_id)

    def update_current_tracks(self, detections: Sequence[Detection]):
        for track in self.tracks:
            if track.state != TrackState.Lost and track.frame_id != self.frame_id:
                track.state = TrackState.Lost

        for detection in detections:
            if not detection.matched:
                track = Track(max_history=self.config.max_history)
                detection.matched = True
                track.update(detection, self.frame_id, self.next_id())
                self.tracks.append(track)

        for track in self.tracks:
            if self.frame_id - track.frame_id > self.config.max_time_lost:
                track.state = TrackState.Removed

        self.tracks = [
            track for track in self.tracks if track.state != TrackState.Removed
        ]

    def run(self, detections: Sequence[Detection]) -> Sequence[Detection]:
        self.frame_id += 1

        for detection in detections:
            detection.matched = False

        self.associate_confirmed_tracks(detections)
        self.associate_unconfirmed_tracks(detections)
        self.update_current_tracks(detections)

        return [track for track in self.tracks if len(track) >= self.config.min_history]



# pyright: reportShadowedImports=false
from collections import deque
from pathlib import Path
from typing import *

import numpy as np

from ...common import Classification, Image, Images, InputFormat
from ...runner import Runner, RunnerConfig
from ..onnx import OnnxNodeInfo, OnnxRunner, OnnxRunnerConfig, OnnxNodeType
from .postproc import TSMPostProcessRunner, TSMPostProcessRunnerConfig
from .preproc import TSMPreProcessRunner, TSMPreProcessRunnerConfig

__all__ = ["TSMModelRunner", "TSMModelRunnerConfig"]


class TSMModelRunnerConfig(
    TSMPreProcessRunnerConfig,
    OnnxRunnerConfig,
    TSMPostProcessRunnerConfig,
    dynamic=True,
):
    type: str = "TSMModelRunner"
    image_node: str = "images"
    score_node: str = "scores"
    batch_size: int = -1


class TSMModelRunner(Runner):
    config_type: Type[RunnerConfig] = TSMModelRunnerConfig

    def prepare(self) -> "TSMModelRunner":
        config: TSMModelRunnerConfig = self.config

        self.onnx_runner = OnnxRunner(config)
        self.onnx_runner.prepare()

        self.input_info: OnnxNodeInfo = self.onnx_runner.metadata[config.image_node]
        self.output_info: OnnxNodeInfo = self.onnx_runner.metadata[config.score_node]

        self.input_buffers: List[OnnxNodeInfo] = []
        self.output_buffers: List[OnnxNodeInfo] = []

        nodes_name = list(self.onnx_runner.metadata.keys())
        nodes_name.sort()

        for name in nodes_name:
            if name == config.image_node or name == config.score_node:
                continue

            node_info: OnnxNodeInfo = self.onnx_runner.metadata[name]
            if node_info.ntype == OnnxNodeType.input:
                self.input_buffers.append(node_info)

            if node_info.ntype == OnnxNodeType.output:
                self.output_buffers.append(node_info)

        self.shift_buffers: Deque[List[np.ndarray]] = deque(
            maxlen=config.max_segments * config.segment_size
        )
        self.zero_buffers: List[np.ndarray] = []
        for node_info in self.input_buffers:
            buffer = np.zeros(node_info.dims, dtype=node_info.dtype)
            self.zero_buffers.append(buffer)

        if config.input_format == InputFormat.hwc:
            (
                config.batch_size,
                config.height,
                config.width,
                _,
            ) = self.input_info.dims

        if config.input_format == InputFormat.chw:
            (
                config.batch_size,
                _,
                config.height,
                config.width,
            ) = self.input_info.dims

        assert config.batch_size == 1, f"Invalid batch size: {config.batch_size}"

        self.preproc_runner = TSMPreProcessRunner(config)
        self.preproc_runner.prepare()

        self.postproc_runner = TSMPostProcessRunner(config)
        self.postproc_runner.prepare()

        return self

    def reset(self):
        self.shift_buffers.clear()
        self.preproc_runner.reset()
        self.postproc_runner.reset()

    def run(
        self,
        images: Union[Image, Images],
        tensor_save_path: str = None,
        tensor_save_txt: Path = None,
        **kwargs,
    ) -> Union[List[Classification], List[List[Classification]]]:
        config: TSMModelRunnerConfig = self.config

        is_single_image = type(images) is not list
        if is_single_image:
            images = [images]

        images = [
            Image(mat=image) if isinstance(image, np.ndarray) else image
            for image in images
        ]

        preproc_images = []
        for image in images:
            preproc_image = self.preproc_runner.run(image)
            mat = preproc_image.mat.astype(self.input_info.dtype)
            preproc_images.append(mat)

        results = []
        for i in range(0, len(preproc_images), config.batch_size):
            batch = preproc_images[i : i + config.batch_size]
            batch_size = len(batch)
            if batch_size < config.batch_size:
                batch += [
                    np.zeros(self.input_info.dims[1:], dtype=self.input_info.dtype)
                    for _ in range(config.batch_size - batch_size)
                ]

            batch = np.stack(batch)

            inputs = {
                config.image_node: batch,
            }

            next_buffers = None
            if len(self.shift_buffers) >= config.step:
                next_buffers = self.shift_buffers[-config.step]
            else:
                next_buffers = self.zero_buffers

            for node_info, buffer in zip(self.input_buffers, next_buffers):
                inputs[node_info.name] = buffer

            if tensor_save_path:
                image_path = tensor_save_path.format(node_name=config.image_node)
                image_path = Path(image_path)
                image_path.parent.mkdir(parents=True, exist_ok=True)
                inputs[config.image_node].tofile(image_path.as_posix())

                paths = [image_path]
                for node_info in self.input_buffers:
                    node_path = tensor_save_path.format(node_name=node_info.name)
                    node_path = Path(node_path)
                    inputs[node_info.name].tofile(node_path.as_posix())
                    paths.append(node_path)

                if tensor_save_txt:
                    with open(tensor_save_txt, "a") as fout:
                        paths = [
                            path.relative_to(tensor_save_txt.parent).as_posix()
                            for path in paths
                        ]
                        fout.write(",".join(paths) + "\n")

            outputs = self.onnx_runner.run(**inputs)

            output = outputs[config.score_node]
            output_buffer = [
                outputs[node_info.name] for node_info in self.output_buffers
            ]
            self.shift_buffers.append(output_buffer)

            output = output[:batch_size, ...]

            classifications = self.postproc_runner.run(output=output)
            results.extend(classifications)

        if is_single_image:
            return results[0]

        return results



# pyright: reportShadowedImports=false
from typing import *

import cv2
import numpy as np

from ...common import InputFormat, ColorFormat, Image, PaddingFormat
from ...runner import RunnerConfig, Runner

__all__ = ["ImagePreProcessRunner", "ImagePreProcessRunnerConfig"]


class ImagePreProcessRunnerConfig(RunnerConfig, dynamic=True):
    type: str = "ImagePreProcessRunner"
    width: int
    height: int
    keep_ratio: bool = False
    interpolation: int = cv2.INTER_LINEAR

    padding_format: PaddingFormat = PaddingFormat.corner
    input_format: InputFormat = InputFormat.hwc
    color_format: ColorFormat = ColorFormat.bgr

    mean: Tuple[float, float, float] = (0, 0, 0)
    std: Tuple[float, float, float] = (1, 1, 1)


class ImagePreProcessRunner(Runner):
    config_type: Type[RunnerConfig] = ImagePreProcessRunnerConfig
    outputs: List[str] = ["output"]

    def prepare(self) -> "ImagePreProcessRunner":
        return self

    def run(self, image: Image, **kwargs) -> Image:
        config: ImagePreProcessRunnerConfig = self.config
        if isinstance(image, np.ndarray):
            image = Image(mat=image)

        mat = image.mat

        resized_height = config.height
        resized_width = config.width

        padding = (0, 0, 0, 0)
        if config.keep_ratio:
            height, width, _ = mat.shape
            tar_ratio = config.width / config.height
            src_ratio = width / height

            if src_ratio > tar_ratio:
                resized_height = int(config.width / src_ratio)

            if src_ratio < tar_ratio:
                resized_width = int(config.height * tar_ratio)

            if config.padding_format == PaddingFormat.corner:
                padding = (
                    0,
                    config.height - resized_height,
                    0,
                    config.width - resized_width,
                )
            elif config.padding_format == PaddingFormat.center:
                pad_width = (config.width - resized_width) // 2
                pad_height = (config.height - resized_height) // 2
                padding = (
                    pad_height,
                    config.height - resized_height - pad_height,
                    pad_width,
                    config.width - resized_width - pad_width,
                )
            else:
                raise ValueError(config.padding_format)

        resized = cv2.resize(
            mat, (resized_width, resized_height), interpolation=config.interpolation
        )

        resized = cv2.copyMakeBorder(
            resized,
            *padding,
            borderType=cv2.BORDER_CONSTANT,
            value=(0, 0, 0),
        )

        padding = (
            padding[0] / config.height,
            padding[1] / config.height,
            padding[2] / config.width,
            padding[3] / config.width,
        )

        if config.color_format == ColorFormat.rgb:
            resized = cv2.cvtColor(resized, cv2.COLOR_BGR2RGB)

        resized = resized.astype(float)
        resized -= config.mean
        resized /= config.std

        if config.input_format == InputFormat.chw:
            resized = resized.transpose(2, 0, 1)

        resized = Image(mat=resized)
        resized.padding = padding

        return resized



# pyright: reportShadowedImports=false
import inspect
from importlib import import_module
from pathlib import Path
from typing import *

from mousse import get_logger

from .cli import Command
from .runner import Runner, RunnerConfig

logger = get_logger(__name__.split(".")[0])

__all__ = ["scan"]


def scan(package: str):
    curr_frame = inspect.currentframe()

    caller = curr_frame.f_back
    filepath = Path(caller.f_code.co_filename).resolve()
    module_path = filepath.parent

    modules = {}
    for child in module_path.glob("*"):
        if child.is_dir() and not child.name.startswith("_"):
            module = import_module(f".{child.name}", package=package)
            modules[child.name] = module

            predicate = lambda val: (
                inspect.isclass(val) and issubclass(val, (Runner, RunnerConfig))
            ) or isinstance(val, Command)
            for key, val in inspect.getmembers(module, predicate=predicate):
                modules[key] = val

    outer_frames = inspect.getouterframes(curr_frame)
    call_frame = outer_frames[1]
    _globals = call_frame.frame.f_globals
    if "__all__" not in _globals:
        _globals["__all__"] = []

    __all = _globals["__all__"]

    if call_frame.function == "<module>":
        __locals = call_frame.frame.f_locals
    else:
        __locals = vars(inspect.getmodule(call_frame.frame))

    for key, val in modules.items():
        __all.append(key)
        __locals[key] = val



# pyright: reportShadowedImports=false
from typing import *

import numpy as np

from ...common import Classification
from ...runner import Runner, RunnerConfig

__all__ = ["ICLSPostProcessRunner", "ICLSPostProcessRunnerConfig"]


class ICLSPostProcessRunnerConfig(RunnerConfig, dynamic=True):
    type: str = "ICLSPostProcessRunnerConfig"
    classes: List[str]


class ICLSPostProcessRunner(Runner):
    config_type: Type[RunnerConfig] = ICLSPostProcessRunnerConfig
    outputs: List[str] = ["classifications"]

    def prepare(self) -> "ICLSPostProcessRunner":
        return self

    def run(self, output: np.ndarray, **kwargs) -> List[Classification]:
        config: ICLSPostProcessRunnerConfig = self.config

        _, num_classes = output.shape
        assert num_classes == len(config.classes), f"Missing classes"

        classifications = []
        for i, scores in enumerate(output):
            scores = output[i, ...].tolist()
            classifications.append(Classification(scores=scores, labels=config.classes))

        return classifications



# pyright: reportShadowedImports=false
from typing import *

from ...runner import RunnerConfig
from ..preproc import ImagePreProcessRunner, ImagePreProcessRunnerConfig

__all__ = ["Yolov5FacePreProcessRunner", "Yolov5FacePreProcessRunnerConfig"]


class Yolov5FacePreProcessRunnerConfig(ImagePreProcessRunnerConfig, dynamic=True):
    type: str = "Yolov5FacePreProcessRunner"
    keep_ratio: bool = True


class Yolov5FacePreProcessRunner(ImagePreProcessRunner):
    config_type: Type[RunnerConfig] = Yolov5FacePreProcessRunnerConfig

    def prepare(self) -> "Yolov5FacePreProcessRunner":
        return self



# pyright: reportShadowedImports=false
from typing import *

from ...runner import RunnerConfig
from ..preproc import ImagePreProcessRunner, ImagePreProcessRunnerConfig

__all__ = ["ICLSPreProcessRunner", "ICLSPreProcessRunnerConfig"]


class ICLSPreProcessRunnerConfig(ImagePreProcessRunnerConfig, dynamic=True):
    type: str = "ICLSPreProcessRunner"


class ICLSPreProcessRunner(ImagePreProcessRunner):
    config_type: Type[RunnerConfig] = ICLSPreProcessRunnerConfig

    def prepare(self) -> "ICLSPreProcessRunner":
        return self



# pyright: reportShadowedImports=false
from collections import OrderedDict
from enum import Enum
from typing import *
from mousse.types.dataclass import Dataclass

import numpy as np
import onnxruntime as ort

import onnx
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE

from ...runner import Runner, RunnerConfig

__all__ = ["OnnxRunner", "OnnxRunnerConfig", "OnnxNodeInfo", "OnnxNodeType"]

def change_batch_size(model, batch_size: int = -1):
    if batch_size > 0:
        for input in model.graph.input:
            dim = input.type.tensor_type.shape.dim[0]
            dim.dim_value = batch_size

        for input in model.graph.input:
            dim = input.type.tensor_type.shape.dim[0]
            dim.dim_value = batch_size

        for output in model.graph.output:
            dim = output.type.tensor_type.shape.dim[0]
            dim.dim_value = batch_size

class OnnxNodeType(str, Enum):
    input: str = "input"
    output: str = "output"


class OnnxNodeInfo(Dataclass, dynamic=True):
    name: str
    dims: Tuple[int, ...]
    dtype: Any
    ntype: OnnxNodeType


class ProviderConfig(Dataclass, dynamic=True):
    type: str = "CPUExecutionProvider"
    params: Dict[str, Any] = {}


class OnnxRunnerConfig(RunnerConfig, dynamic=True):
    type: str = "OnnxRunner"
    path: str
    batch_size: int = -1
    providers: List[ProviderConfig] = [ProviderConfig()]


class OnnxRunner(Runner):
    config_type: Type[RunnerConfig] = OnnxRunnerConfig

    @property
    def outputs(self) -> List[str]:
        return [output_node.name for output_node in self.model.graph.output]

    def prepare(self) -> "OnnxRunner":
        config: OnnxRunnerConfig = self.config
        self.model = onnx.load(config.path)
        
        if config.batch_size > 0:
            change_batch_size(self.model, config.batch_size)
        
        self.nodes = OrderedDict()
        
        providers = []
        for provider in config.providers:
            if provider.params:
                providers.append((provider.type, provider.params))
                continue
            providers.append(provider.type)

        self.sess = ort.InferenceSession(
            config.path,
            providers=providers,
        )

        self.metadata.clear()
        for input_node in self.model.graph.input:
            dims = []
            for dim in input_node.type.tensor_type.shape.dim:
                dims.append(dim.dim_value)
            dtype = TENSOR_TYPE_TO_NP_TYPE[input_node.type.tensor_type.elem_type]
            node_info = OnnxNodeInfo(
                name=input_node.name,
                dims=tuple(dims),
                dtype=dtype,
                ntype=OnnxNodeType.input,
            )

            self.nodes[input_node.name] = node_info
            self.metadata[input_node.name] = node_info

        for output_node in self.model.graph.output:
            dims = []
            for dim in output_node.type.tensor_type.shape.dim:
                dims.append(dim.dim_value)
            dtype = TENSOR_TYPE_TO_NP_TYPE[output_node.type.tensor_type.elem_type]
            node_info = OnnxNodeInfo(
                name=output_node.name,
                dims=tuple(dims),
                dtype=dtype,
                ntype=OnnxNodeType.output,
            )

            self.nodes[output_node.name] = node_info
            self.metadata[output_node.name] = node_info

        return self

    def run(self, **inputs: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        for input_node in self.model.graph.input:
            assert input_node.name in inputs, f"{input_node.name} not found"
            dtype = TENSOR_TYPE_TO_NP_TYPE[input_node.type.tensor_type.elem_type]
            inputs[input_node.name] = inputs[input_node.name].astype(dtype)

        output_mats = self.sess.run(None, inputs)
        return {key: val for key, val in zip(self.outputs, output_mats)}



# pyright: reportShadowedImports=false
from enum import Enum
from typing import *

import numpy as np
from mousse import Dataclass


__all__ = ["BBoxType", "BBox", "iou", "ioa"]


class BBoxType(str, Enum):
    ltrb: str = "ltrb"
    ltwh: str = "ltwh"
    xyah: str = "xyah"


def ltrb2ltwh(source: Tuple[float, float, float, float]):
    return (source[0], source[1], source[2] - source[0], source[3] - source[1])


def ltwh2ltrb(source: Tuple[float, float, float, float]):
    return (source[0], source[1], source[2] + source[0], source[3] + source[1])


def ltrb2xyah(source: Tuple[float, float, float, float]):
    return (
        (source[0] + source[2]) / 2,
        (source[1] + source[3]) / 2,
        (source[2] - source[0]) / (source[3] - source[1]),
        source[3] - source[1],
    )


def xyah2ltrb(source: Tuple[float, float, float, float]):
    x, y, aspect, height = source
    width = aspect * height
    return (x - width / 2, y - height / 2, x + width / 2, y + height / 2)


__CONVERTER__ = {
    (BBoxType.ltrb, BBoxType.ltwh): ltrb2ltwh,
    (BBoxType.ltwh, BBoxType.ltrb): ltwh2ltrb,
    (BBoxType.ltrb, BBoxType.xyah): ltrb2xyah,
    (BBoxType.xyah, BBoxType.ltrb): xyah2ltrb,
}


class BBox(Dataclass, dynamic=True):
    rect: Tuple[float, float, float, float]
    type: BBoxType = BBoxType.ltrb
    score: float = -1

    def to(self, btype: BBoxType) -> "BBox":
        if btype == self.type:
            return self

        if (self.type, btype) in __CONVERTER__:
            rect = __CONVERTER__[(self.type, btype)](self.rect)
            return BBox(rect=rect, type=btype, score=self.score)

        for itype in BBoxType:
            if (self.type, itype) in __CONVERTER__ and (itype, btype) in __CONVERTER__:
                rect = __CONVERTER__[(self.type, itype)](self.rect)
                rect = __CONVERTER__[(itype, btype)](rect)
                return BBox(rect=rect, type=btype, score=self.score)

        raise NotImplementedError()

    def asarray(self, dtype: Type[Any] = float) -> np.ndarray:
        return np.asarray(self.astuple(dtype=dtype))

    def astuple(self, dtype: Type[Any] = float) -> Tuple[float, float, float, float]:
        return tuple(dtype(val) for val in self.rect)

    @property
    def valid(self):
        if self.type == BBoxType.ltrb:
            if self.rect[2] < self.rect[0] or self.rect[3] < self.rect[1]:
                return False

            return True

        bbox = self.to(BBoxType.ltrb)
        return bbox.valid

    @property
    def width(self):
        if not self.valid:
            return 0

        if self.type == BBoxType.ltrb:
            return self.rect[2] - self.rect[0]

        return self.to("ltrb").width

    @property
    def height(self):
        if not self.valid:
            return 0

        if self.type == BBoxType.ltrb:
            return self.rect[3] - self.rect[1]

        return self.to("ltrb").height

    @property
    def area(self):
        return self.width * self.height if self.valid else 0

    def __and__(self, other: "BBox") -> "BBox":
        if other.type != BBoxType.ltrb:
            other = other.to("ltrb")

        curr = self
        if curr.type != BBoxType.ltrb:
            curr = self.to("ltrb")

        left = max(curr.rect[0], other.rect[0])
        top = max(curr.rect[1], other.rect[1])
        right = min(curr.rect[2], other.rect[2])
        bottom = min(curr.rect[3], other.rect[3])

        intersection = BBox(
            rect=[left, top, right, bottom],
            type=BBoxType.ltrb,
            score=(self.score + other.score) / 2,
        )
        intersection = intersection.to(self.type)
        if not intersection.valid:
            intersection.rect = (-1, -1, -1, -1)

        return intersection


def iou(a: BBox, b: BBox) -> float:
    intersection = a & b
    if intersection.area == 0:
        return 0
    return intersection.area / (a.area + b.area - intersection.area)


def ioa(a: BBox, b: BBox) -> float:
    intersection = a & b
    if intersection.area == 0:
        return 0
    return intersection.area / a.area



# pyright: reportShadowedImports=false
import math
from typing import *

import numpy as np
from mousse import Dataclass

from ...common import Detection, clip
from ...runner import Runner, RunnerConfig

__all__ = ["YoloXPostProcessRunner", "YoloXPostProcessRunnerConfig"]


class YoloXGrid(Dataclass):
    x: int
    y: int
    stride: int


class YoloXPostProcessRunnerConfig(RunnerConfig):
    type: str = "YoloXPostProcessRunner"

    strides: List[int] = []
    conf_threshold: float = 0.5
    nms_threshold: float = 0.3

    height: int = -1
    width: int = -1
    num_anchors: int = -1
    num_classes: int = -1


class YoloxPostProcessInputData(Dataclass):
    images: np.ndarray
    output: np.ndarray


class YoloXPostProcessRunner(Runner):
    config_type = YoloXPostProcessRunnerConfig
    outputs = ["detections"]

    def prepare(self) -> "YoloXPostProcessRunner":
        config: YoloXPostProcessRunnerConfig = self.config

        self.grids: List[YoloXGrid] = []
        if config.width > 0 and config.height > 0:
            self.generate_grids()

        return self

    def generate_grids(self):
        config: YoloXPostProcessRunnerConfig = self.config
        self.grids.clear()

        for stride in config.strides:
            num_grid_w = config.width // stride
            num_grid_h = config.height // stride
            for y in range(num_grid_h):
                for x in range(num_grid_w):
                    self.grids.append(YoloXGrid(x=x, y=y, stride=stride))

        if config.num_anchors > 0:
            assert (
                len(self.grids) == config.num_anchors
            ), f"Number of anchors mismatch, expect {config.num_anchors}, get {len(self.grids)}"

    def generate_proposals(self, output: np.ndarray) -> List[Detection]:
        config: YoloXPostProcessRunnerConfig = self.config
        proposals = []
        for grid_idx, grid in enumerate(self.grids):
            tensor = output[grid_idx, ...]

            x_center = float(tensor[0] + grid.x) * grid.stride
            y_center = float(tensor[1] + grid.y) * grid.stride
            width = math.exp(tensor[2]) * grid.stride
            height = math.exp(tensor[3]) * grid.stride

            objectness = float(tensor[4])

            for class_idx in range(config.num_classes):
                cls_score = tensor[5 + class_idx]
                box_score = objectness * cls_score

                if box_score > config.conf_threshold:
                    left = x_center - width / 2
                    top = y_center - height / 2
                    right = x_center + width / 2
                    bottom = y_center + height / 2
                    proposals.append(
                        Detection(
                            left=left,
                            top=top,
                            right=right,
                            bottom=bottom,
                            label=class_idx,
                            prob=box_score,
                        )
                    )
        return proposals

    def nms(self, proposals: List[Detection]) -> List[Detection]:
        config: Detection = self.config
        selections: List[Detection] = []
        for proposal in proposals:
            keep = True

            area = proposal.width * proposal.height
            for selection in selections:
                left = max(proposal.left, selection.left)
                top = max(proposal.top, selection.top)
                right = min(proposal.right, selection.right)
                bottom = min(proposal.bottom, selection.bottom)

                if right <= left or bottom <= top:
                    continue

                inter_area = (right - left) * (bottom - top)
                union_area = area + selection.width * selection.height - inter_area

                if inter_area / union_area > config.nms_threshold:
                    keep = False
                    break

            if keep:
                selections.append(proposal)

        return selections

    def run(self, output: np.ndarray, **kwargs) -> List[List[Detection]]:
        config: YoloXPostProcessRunnerConfig = self.config

        batch_size, *_ = output.shape

        batch_proposals = []
        for sample_id in range(batch_size):
            sample_output = output[sample_id, ...]
            proposals = self.generate_proposals(sample_output)
            batch_proposals.append(proposals)

        batch_objects = []
        for proposals in batch_proposals:
            proposals.sort(key=lambda proposal: proposal.prob)
            proposals = self.nms(proposals)

            objects = []
            for proposal in proposals:
                proposal.left = clip(proposal.left, 0, config.width)
                proposal.top = clip(proposal.top, 0, config.height)
                proposal.right = clip(proposal.right, 0, config.width)
                proposal.bottom = clip(proposal.bottom, 0, config.height)

                if proposal.right <= proposal.left or proposal.bottom <= proposal.top:
                    continue

                proposal.left /= config.width
                proposal.top /= config.height
                proposal.right /= config.width
                proposal.bottom /= config.height

                objects.append(proposal)

            batch_objects.append(objects)

        return batch_objects



# pyright: reportShadowedImports=false
from collections import deque
from enum import Enum
from typing import *

from mousse import Dataclass, Field

from ...common import BBox, BBoxType, Detection
from .kalman_filter import KalmanFilter

__all__ = ["TrackState", "Track"]


class TrackState(str, Enum):
    New = "new"
    Tracked = "tracked"
    Lost = "lost"
    Removed = "removed"


class Track(Dataclass, dynamic=True):
    max_history: int = 5
    detection: Detection = None

    label: int
    frame_id: int = 0
    track_id: int = 0
    state: TrackState = TrackState.New

    kf: KalmanFilter = Field(factory=KalmanFilter, private=True)

    def __build__(self, **kwargs):
        self.bboxes: Sequence[BBox] = deque([], maxlen=self.max_history)

    @property
    def last_bbox(self) -> BBox:
        assert len(self.bboxes) > 0, "List of bboxes is empty"
        return self.bboxes[-1]

    @property
    def pred_bbox(self) -> BBox:
        bbox = BBox(rect=self.kf.mean[:4], type=BBoxType.xyah)
        return bbox.to("ltrb")

    @property
    def mean_area(self):
        return sum(bbox.area for bbox in self.bboxes) / (len(self.bboxes) or 1)

    @property
    def mean_score(self):
        return sum(bbox.score for bbox in self.bboxes) / (len(self.bboxes) or 1)
    
    def update(self, detection: Detection, frame_id: int, track_id: int = -1, *kwargs):
        self.detection = detection
        
        bbox = self.detection.to("ltrb")
        self.bboxes.append(bbox)
        self.label = self.detection.label

        self.frame_id = frame_id
        if self.track_id > 0:
            self.kf.update(self.last_bbox.to("xyah").astuple())
        else:
            self.kf.initiate(self.last_bbox.to("xyah").astuple())

        
        if track_id > 0:
            self.track_id = track_id
        
        self.state = TrackState.Tracked
        self.detection.track_id = track_id

    def predict(self):
        self.kf.predict()

    def __len__(self):
        return len(self.bboxes)



# pyright: reportShadowedImports=false
from pathlib import Path
from typing import *

import onnx
from onnxsim import simplify
from typer import Argument

from .cli import cli

__all__ = ["clean"]


@cli.command()
def clean(
    source: Path = Argument(..., help="Source onnx"),
    target: Path = Argument(..., help="Target onnx"),
):
    assert (
        source.exists() and source.is_file() and source.suffix.lower() == ".onnx"
    ), f"{source} is invalid"
    model = onnx.load(source.as_posix())
    model_simp, check = simplify(model)

    if check:
        model = model_simp

    if model.ir_version < 4:
        print(
            "Model with ir_version below 4 requires to include initilizer in graph input"
        )
        return

    inputs = model.graph.input
    name_to_input = {}
    for input in inputs:
        name_to_input[input.name] = input

    for initializer in model.graph.initializer:
        if initializer.name in name_to_input:
            inputs.remove(name_to_input[initializer.name])

    target.parent.mkdir(exist_ok=True, parents=True)
    onnx.save(model, target.as_posix())



# pyright: reportShadowedImports=false
import subprocess
from pathlib import Path
from typing import *

import setuptools

PACKAGE = "soboro"
CURRENT_DIR = Path(__file__).resolve().parent

with open(CURRENT_DIR / "README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def read_requirements(path: Union[str, Path]):
    with open(path, "r") as fh:
        return {line.strip() for line in fh.readlines() if not line.startswith("#")}


__VERSION__ = "0.0.4"

requirements = list(read_requirements(CURRENT_DIR / "requirements.txt"))
extras_require = {"all": set()}


packages = setuptools.find_packages()
sub_packages = []
for sub_requirement in (CURRENT_DIR / PACKAGE).rglob("requirements.txt"):
    sub_package = sub_requirement.parent.relative_to(CURRENT_DIR)
    sub_requirements = read_requirements(sub_requirement)
    extras_require[sub_package.name] = sub_requirements
    extras_require["all"].update(sub_requirements)

extras_require = {key: list(val) for key, val in extras_require.items()}
entry_points = {"console_scripts": (f"{PACKAGE} = {PACKAGE}.__main__:main",)}

try:
    subprocess.check_output(["which", "nvidia-smi"])
except subprocess.CalledProcessError:
    pass
else:
    num_gpus = str(subprocess.check_output(["nvidia-smi", "-L"])).count('UUID')
    if num_gpus:
        requirements.remove("onnxruntime")
        requirements.append("onnxruntime-gpu")
        for values in extras_require.values():
            if "onnxruntime" in values:
                values.remove("onnxruntime")
                values.append("onnxruntime-gpu")

setuptools.setup(
    name=PACKAGE,
    packages=packages,
    version=__VERSION__,
    author="datnh21",
    author_email="v.datnh21@vinai.io",
    description=PACKAGE,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    entry_points=entry_points,
    install_requires=requirements,
    extras_require=extras_require,
)



# pyright: reportShadowedImports=false
from enum import Enum

__all__ = ["MonitorStatus"]


class MonitorStatus(str, Enum):
    unknown: str = "unknown"
    warned: str = "warned"
    warning: str = "warning"
    checking: str = "checking"



# pyright: reportShadowedImports=false
from typing import *

import numpy as np
from .types import Detection

__all__ = ["clip", "sigmoid", "nms", "softmax"]


def clip(val: float, lower: float, upper: float):
    return lower if val < lower else upper if val > upper else val


def sigmoid(x: Union[np.ndarray, float]):
    return 1 / (1 + np.exp(-x))


def softmax(x: np.ndarray, axis: int = -1):
    """Compute softmax values for each sets of scores in x."""
    sum_score = np.sum(np.exp(x), axis=axis)
    sum_score = sum_score[:, np.newaxis]
    return np.exp(x) / sum_score


def nms(proposals: List[Detection], nms_threshold: float) -> List[Detection]:
    selections: List[Detection] = []
    for proposal in proposals:
        keep = True

        area = proposal.width * proposal.height
        for selection in selections:
            left = max(proposal.left, selection.left)
            top = max(proposal.top, selection.top)
            right = min(proposal.right, selection.right)
            bottom = min(proposal.bottom, selection.bottom)

            if right <= left or bottom <= top:
                continue

            inter_area = (right - left) * (bottom - top)
            union_area = area + selection.width * selection.height - inter_area

            if inter_area / union_area > nms_threshold:
                keep = False
                break

        if keep:
            selections.append(proposal)

    return selections



def main():
    from mousse import get_logger
    from .cli import cli

    package_name, *_ = __name__.split(".")
    logger = get_logger(package_name)
    logger.include_extra = False
    logger.add_handler("RotatingFileHandler", path=f"logs/{package_name}.out")

    cli()


if __name__ == "__main__":
    main()



# pyright: reportShadowedImports=false
from typing import *

from ...runner import RunnerConfig
from ..preproc import ImagePreProcessRunner, ImagePreProcessRunnerConfig

__all__ = ["TSMPreProcessRunner", "TSMPreProcessRunnerConfig"]


class TSMPreProcessRunnerConfig(ImagePreProcessRunnerConfig, dynamic=True):
    type: str = "Yolov5FacePreProcessRunner"
    keep_ratio: bool = True


class TSMPreProcessRunner(ImagePreProcessRunner):
    config_type: Type[RunnerConfig] = TSMPreProcessRunnerConfig

    def prepare(self) -> "TSMPreProcessRunnerConfig":
        return self



# pyright: reportShadowedImports=false
import math
from typing import *

import numpy as np

from ...common import FaceDetection, Image, Images, InputFormat, clip
from ...runner import Runner, RunnerConfig
from ..onnx import OnnxNodeType, OnnxRunner, OnnxRunnerConfig
from .postproc import Yolov5FacePostProcessRunner, Yolov5FacePostProcessRunnerConfig
from .preproc import Yolov5FacePreProcessRunner, Yolov5FacePreProcessRunnerConfig

__all__ = ["Yolov5FaceModelRunner", "Yolov5FaceModelRunnerConfig"]


class Yolov5FaceModelRunnerConfig(
    Yolov5FacePreProcessRunnerConfig,
    OnnxRunnerConfig,
    Yolov5FacePostProcessRunnerConfig,
    dynamic=True,
):
    type: str = "Yolov5FaceModelRunner"
    batch_size: int = -1


class Yolov5FaceModelRunner(Runner):
    config_type: Type[RunnerConfig] = Yolov5FaceModelRunnerConfig
    outputs: List[str] = ["detections"]

    def prepare(self) -> "Yolov5FaceModelRunner":
        config: Yolov5FaceModelRunnerConfig = self.config

        self.onnx_runner = OnnxRunner(config)
        self.onnx_runner.prepare()

        for node_name, node_info in self.onnx_runner.metadata.items():
            if node_info.ntype == OnnxNodeType.input:
                self.input_info = node_info
                if config.input_format == InputFormat.hwc:
                    (
                        config.batch_size,
                        config.height,
                        config.width,
                        _,
                    ) = self.input_info.dims

                if config.input_format == InputFormat.chw:
                    (
                        config.batch_size,
                        _,
                        config.height,
                        config.width,
                    ) = self.input_info.dims

            if node_info.ntype == OnnxNodeType.output:
                config.outputs[node_name] = tuple(node_info.dims)

        self.preproc_runner = Yolov5FacePreProcessRunner(config)
        self.preproc_runner.prepare()

        self.postproc_runner = Yolov5FacePostProcessRunner(config)
        self.postproc_runner.prepare()

        return self

    def run(
        self, images: Union[Image, Images], **kwargs
    ) -> Union[List[FaceDetection], List[List[FaceDetection]]]:
        config: Yolov5FaceModelRunnerConfig = self.config

        is_single_image = type(images) is not list
        if is_single_image:
            images = [images]

        images = [
            Image(mat=image) if isinstance(image, np.ndarray) else image
            for image in images
        ]

        preproc_images = []
        for image in images:
            preproc_image = self.preproc_runner.run(image)
            preproc_images.append(preproc_image.mat)

        candidates = []
        for i in range(0, len(preproc_images), config.batch_size):
            batch = preproc_images[i : i + config.batch_size]
            batch_size = len(batch)
            if batch_size < config.batch_size:
                batch += [
                    np.zeros(self.input_info.dims[1:], dtype=self.input_info.dtype)
                    for _ in range(config.batch_size - batch_size)
                ]

            batch = np.stack(batch)
            outputs = self.onnx_runner.run(**{self.input_info.name: batch})

            batch = batch[:batch_size, ...]
            outputs = {key: val[:batch_size, ...] for key, val in outputs.items()}

            detections = self.postproc_runner.run(**outputs)
            candidates.extend(detections)

        results = []
        for candidate, image in zip(candidates, images):
            height, width, _ = image.mat.shape
            top, bottom, left, right = image.padding

            result = []
            for detection in candidate:
                detection.left = clip(detection.left - left, 0, 1)
                detection.top = clip(detection.top - top, 0, 1)
                detection.right = clip(detection.right - right, 0, 1)
                detection.bottom = clip(detection.bottom - bottom, 0, 1)

                if (
                    detection.right <= detection.left
                    or detection.bottom <= detection.top
                ):
                    continue

                detection.left = math.floor(detection.left * width)
                detection.top = math.floor(detection.top * height)
                detection.right = math.floor(detection.right * width)
                detection.bottom = math.floor(detection.bottom * height)

                for point in detection.keypoints:
                    point.x *= width
                    point.y *= height

                result.append(detection)

        results.append(result)

        if is_single_image:
            return results[0]

        return results



# pyright: reportShadowedImports=false
from typing import *

from ...common import Detection, Image, Images
from ...runner import Runner, RunnerConfig

__all__ = ["ODACRunnerConfig", "ODACRunner"]


class ODACRunnerConfig(RunnerConfig, dynamic=True):
    type: str = "ODACRunner"
    detector: RunnerConfig
    tracker: RunnerConfig = None


class ODACRunner(Runner):
    config_type: Type[RunnerConfig] = ODACRunnerConfig
    outputs: List[str] = ["detections"]

    def prepare(self) -> "Runner":
        config: ODACRunnerConfig = self.config

        self.detector = Runner.create(config.detector)
        self.detector.prepare()

        self.tracker = None
        if config.tracker is not None:
            self.tracker = Runner.create(config.tracker)
            self.tracker.prepare()

    def run(
        self, images: Union[Image, Images], **kwargs
    ) -> Union[List[Detection], List[List[Detection]]]:
        is_single_image = type(images) is not list
        if is_single_image:
            images = [images]

        images_detection = self.detector.run(images)
        results = []
        for detections in images_detection:
            if self.tracker:
                tracks = self.tracker.run(detections=detections)
                results.append([track.detection for track in tracks])
                continue

            results.append(detections)

        if is_single_image:
            return results[0]

        return results



# pyright: reportShadowedImports=false
from enum import Enum
from typing import *

import numpy as np
from mousse import Dataclass


__all__ = ["Image", "Images", "InputFormat", "ColorFormat", "PaddingFormat"]

Number = Union[int, float]


class Image(Dataclass, dynamic=True):
    mat: np.ndarray
    padding: Tuple[Number, Number, Number, Number] = (0, 0, 0, 0)


Images = List[Image]


class InputFormat(str, Enum):
    hwc: str = "hwc"
    chw: str = "chw"


class ColorFormat(str, Enum):
    rgb: str = "rgb"
    bgr: str = "bgr"
    yuv: str = "yuv"


class PaddingFormat(str, Enum):
    corner: str = "corner"
    center: str = "center"
